private import antlrrt.runtime;
private import antlrrt.utils;

private import std.conv;



alias asWStringT!(int) asWString2;




public abstract class CalcAST : BaseAST {
	public abstract int value();
	
	// satisfy abstract methods from BaseAST
	public void initialize(int t, WString txt) {
	}
	public void initialize(AST t) {
	}
	public void initialize(Token tok) {
	}
}




/** A simple node to represent an INT */
public class INTNode : CalcAST {
	int v=0;

    public static AST antlrFactory()
    {
        return new INTNode();
    }

	public this() {
	}

	public this(Token tok) {
    	initialize(tok);
	}
    	
	public void initialize(Token tok) {
		v = toInt(tok.getText().toString());
	}

	/** Compute value of subtree; this is heterogeneous part :) */
	public int value() {
		return v;
	}

	public WString toWString() {
		return " " ~ asWString2(v);
	}

	public void xmlSerializeNode(Writer outp) {
		outp.write("<int>" ~ asWString2(v) ~ "</int>");
	}
}



public abstract class BinaryOperatorAST : CalcAST {
	public CalcAST left() {
		return cast(CalcAST)getFirstChild();
	}

	public CalcAST right() {
		CalcAST t = left();
		if ( t is null ) return null;
		return cast(CalcAST)t.getNextSibling();
	}
}




/** A simple node to represent MULT operation */
public class MULTNode : BinaryOperatorAST {
	public this(Token tok) {
	}
	
    public static AST antlrFactory()
    {
        return new MULTNode(null);
    }


	/** Compute value of subtree; this is heterogeneous part :) */
	public int value() {
		return left().value() * right().value();
	}

	public WString toWString() {
		return asWString(" *");
	}

	public void xmlSerializeRootOpen(Writer outp) {
		outp.write(asWString("<MULT>"));
	}

	public void xmlSerializeRootClose(Writer outp) {
		outp.write(asWString("</MULT>"));
	}
}



/** A simple node to represent PLUS operation */
public class PLUSNode : BinaryOperatorAST {
	public this(Token tok) {
	}
	
	public static AST antlrFactory()
    {
        return new PLUSNode(null);
    }

	/** Compute value of subtree; this is heterogeneous part :) */
	public int value() {
		return left().value() + right().value();
	}

	public WString toWString() {
		return asWString(" +");
	}

	public void xmlSerializeRootOpen(Writer outp) {
		outp.write(asWString("<PLUS>"));
	}

	public void xmlSerializeRootClose(Writer outp) {
		outp.write(asWString("</PLUS>"));
	}
}


