/** Various utilities for the modules (including some
  faked java classes)
*/
module antlrrt.collections;

private import antlrrt.utils;

// private import std.stream;
// private import std.stdio;

// alias std.stream.InputStream InputStream;

// private import antlrrt.exceptions;

/** Parametrized Java fake class Hashtable (an AA wrapper).
    @param K Key type
    @param V Value type
*/
class Hashtable(K, V)
{
    V[K] data;
    
    void put(K k, V v)
    {
        data[k] = v;
    }
    
    /// Get pointer to data
    V* getptr(K k)
    {
        return k in data;
    }
    
    
    V get(K k)
    {
        V* v = getptr(k);
        if (!v)
          return cast(V) null;   // TODO: Always right?
        else
          return *v;
    }
    
    void remove(K k)
    {
        data.remove(k);
    }
}


/** A simple stack definition; restrictive in that you cannot
 * access arbitrary stack elements.
 *
 * @author Terence Parr
 * <a href=http://www.MageLang.com>MageLang Institute</a>
 */
public interface Stack(T) {
    public int height();

    public T pop();  // throws NoSuchElementException;

    public void push(T e);

    public T top();   // throws NoSuchElementException;
}



/** A simple parametrized Vector
*/
// TODO: Refine
class Vector(T): Stack!(T)
{
    protected T[] data;


    public this()
    {}
    
    public this(int cap) {
        this();
    }
    
    public void appendElement(T e) {
        data ~= e;
    }
    
    public T get(int i) {
        return data[i];
    }
    
    public T elementAt(int i) {
        return get(i);
    }
    
    public int length()
    {
        return data.length;
    }
    
    public void drop()
    {
        if (data.length > 0)
            data.length = data.length - 1;
    }
    
    // Stack interface:
    public int height()
    {
        return length();
    }

    public T pop()  // throws NoSuchElementException;
    {
        int l = length();
        if (l == 0)
            throw new NoSuchElementException(asWString("in Vector!(T).pop"));
            
        T result = get(l - 1);
        drop();
        
        return result;
    }

    public void push(T e)
    {
        appendElement(e);
    }

    public T top()   // throws NoSuchElementException;
    {
        int l = length();
        if (l == 0)
            throw new NoSuchElementException(asWString("in Vector!(T).top"));
            
        return get(l - 1);
    }
}
    

/** A queue implementation spezialized to CharQueue and TokenQueue
*/
public class Queue(T) {
    /** Physical circular buffer of elements */
    protected T[] buffer;
    /** buffer.length-1 for quick modulos */
    private int sizeLessOne;
    /** physical index of front elements */
    private int offset;
    /** number of element in the queue */
    public int nbrEntries;

    public this(int minSize) {
        // Find first power of 2 >= to requested size
        int size;
        if ( minSize<0 ) {
            init(16); // pick some value for them
            return;
        }
        // check for overflow
        if ( minSize>=(int.max/2) ) {
            init(int.max); // wow that's big.
            return;
        }
        for (size = 2; size < minSize; size *= 2) {
        }
        init(size);
    }

    /** Add token to end of the queue
     * @param tok The element to add
     */
    public final void append(T tok) {
        if (nbrEntries == buffer.length) {
            expand();
        }
        buffer[(offset + nbrEntries) & sizeLessOne] = tok;
        nbrEntries++;
    }

    /** Fetch an element from the queue by index
     * @param idx The index of the element to fetch, where zero is the element at the front of the queue
     */
    public final T elementAt(int idx) {
        return buffer[(offset + idx) & sizeLessOne];
    }

    /** Expand the token buffer by doubling its capacity */
    private final void expand() {
        T[] newBuffer = new T[buffer.length * 2];
        // Copy the contents to the new buffer
        // Note that this will store the first logical item in the
        // first physical array element.
        for (int i = 0; i < buffer.length; i++) {
            newBuffer[i] = elementAt(i);
        }
        // Re-initialize with new contents, keep old nbrEntries
        buffer = newBuffer;
        sizeLessOne = buffer.length - 1;
        offset = 0;
    }

    /** Initialize the queue.
     * @param size The initial size of the queue
     */
    public void init(int size) {
        // Allocate buffer
        buffer = new T[size];
        // Other initialization
        sizeLessOne = size - 1;
        offset = 0;
        nbrEntries = 0;
    }

    /** Clear the queue. Leaving the previous buffer alone.
     */
    public final void reset() {
        offset = 0;
        nbrEntries = 0;
    }

    /** Remove element from front of queue */
    public final void removeFirst() {
        offset = (offset + 1) & sizeLessOne;
        nbrEntries--;
    }
}



/** Interface used by BitSet to format elements of the set when
 * converting to string
 */
public interface CharFormatter {

    public WString escapeChar(int c, boolean forCharLiteral);

    public WString escapeString(WString s);

    public WString literalChar(int c);

    public WString literalString(WString s);
}


/**A BitSet to replace java.util.BitSet.
 * Primary differences are that most set operators return new sets
 * as opposed to oring and anding "in place".  Further, a number of
 * operations were added.  I cannot contain a BitSet because there
 * is no way to access the internal bits (which I need for speed)
 * and, because it is final, I cannot subclass to add functionality.
 * Consider defining set degree.  Without access to the bits, I must
 * call a method n times to test the ith bit...ack!
 *
 * Also seems like or() from util is wrong when size of incoming set is bigger
 * than this.bits.length.
 *
 * @author Terence Parr
 * @author <br><a href="mailto:pete@yamuna.demon.co.uk">Pete Wells</a>
 */
public class BitSet {    // implements Cloneable
    protected const static int BITS = 64;    // number of bits / long
    protected const static int NIBBLE = 4;
    protected const static int LOG_BITS = 6; // 2^6 == 64

    /* We will often need to do a mod operator (i mod nbits).  Its
     * turns out that, for powers of two, this mod operation is
     * same as (i & (nbits-1)).  Since mod is slow, we use a
     * precomputed mod mask to do the mod instead.
     */
    protected const static int MOD_MASK = BITS - 1;

    /** The actual data bits */
    protected long bits[];

    /** Construct a bitset of size one word (64 bits) */
    public this() {
        this(BITS);
    }

    /** Construction from a static array of longs */
    public this(long[] bits_) {
        bits = bits_;
    }

    /** Construct a bitset given the size
     * @param nbits The size of the bitset in bits
     */
    public this(int nbits) {
        bits = new long[((nbits - 1) >> LOG_BITS) + 1];
    }

    /** or this element into this set (grow as necessary to accommodate) */
    public void add(int el) {
        //System.outp.println("add("+el+")");
        int n = wordNumber(el);
        //System.outp.println("word number is "+n);
        //System.outp.println("bits.length "+bits.length);
        if (n >= bits.length) {
            growToInclude(el);
        }
        bits[n] |= bitMask(el);
    }

    public BitSet and(BitSet a) {
        BitSet s = this.clone();
        s.andInPlace(a);
        return s;
    }

    public void andInPlace(BitSet a) {
        int min = Math.min(bits.length, a.bits.length);
        for (int i = min - 1; i >= 0; i--) {
            bits[i] &= a.bits[i];
        }
        // clear all bits in this not present in a (if this bigger than a).
        for (int i = min; i < bits.length; i++) {
            bits[i] = 0;
        }
    }

    private final static long bitMask(int bitNumber) {
        int bitPosition = bitNumber & MOD_MASK; // bitNumber mod BITS
        return 1L << bitPosition;
    }

    public void clear() {
        for (int i = bits.length - 1; i >= 0; i--) {
            bits[i] = 0;
        }
    }

    public void clear(int el) {
        int n = wordNumber(el);
        if (n >= bits.length) {    // grow as necessary to accommodate
            growToInclude(el);
        }
        bits[n] &= ~bitMask(el);
    }

/*
    public Object clone() {
        BitSet s;
        try {
            s = (BitSet)super.clone();
            s.bits = new long[bits.length];
            System.arraycopy(bits, 0, s.bits, 0, bits.length);
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
        return s;
    }
*/

    public BitSet clone() {
        return new BitSet(bits.dup);
    }
    

    public int degree() {
        int deg = 0;
        for (int i = bits.length - 1; i >= 0; i--) {
            long word = bits[i];
            if (word != 0L) {
                for (int abit = BITS - 1; abit >= 0; abit--) {
                    if ((word & (1L << abit)) != 0) {
                        deg++;
                    }
                }
            }
        }
        return deg;
    }

    /** code "inherited" from java.util.BitSet */
    public boolean equals(Object obj) {
        BitSet set = cast(BitSet) obj;
        if (set !is null) {
//        if ((obj != null) && (obj instanceof BitSet)) {
            // BitSet set = (BitSet)obj;

            int n = Math.min(bits.length, set.bits.length);
            for (int i = n; i-- > 0;) {
                if (bits[i] != set.bits[i]) {
                    return false;
                }
            }
            if (bits.length > n) {
                for (int i = bits.length; i-- > n;) {
                    if (bits[i] != 0) {
                        return false;
                    }
                }
            }
            else if (set.bits.length > n) {
                for (int i = set.bits.length; i-- > n;) {
                    if (set.bits[i] != 0) {
                        return false;
                    }
                }
            }
            return true;
        }
        return false;
    }

    /+ +++++ // TODO Unused?
    /** Find ranges in a set element array.  @param elems The array of
     * elements representing the set, usually from Bit Set.toArray().
     * @return Vector of ranges.
     */
    public static Vector getRanges(int[] elems) {
        if (elems.length == 0) {
            return null;
        }
        int begin = elems[0];
        int end = elems[elems.length - 1];
        if (elems.length <= 2) {
            // Not enough elements for a range expression
            return null;
        }

        Vector ranges = new Vector(5);
        // look for ranges
        for (int i = 0; i < elems.length - 2; i++) {
            int lastInRange;
            lastInRange = elems.length - 1;
            for (int j = i + 1; j < elems.length; j++) {
                if (elems[j] != elems[j - 1] + 1) {
                    lastInRange = j - 1;
                    break;
                }
            }
            // found a range
            if (lastInRange - i > 2) {
                ranges.appendElement(new IntRange(elems[i], elems[lastInRange]));
            }
        }
        return ranges;
    }
    +++++ +/

    /**
     * Grows the set to a larger number of bits.
     * @param abit element that must fit in set
     */
    public void growToInclude(int abit) {
        int newSize = Math.max(bits.length << 1, numWordsToHold(abit));
        setSize(newSize);
    }

    public boolean member(int el) {
        int n = wordNumber(el);
        if (n >= bits.length) return false;
        return (bits[n] & bitMask(el)) != 0;
    }

    public boolean nil() {
        for (int i = bits.length - 1; i >= 0; i--) {
            if (bits[i] != 0) return false;
        }
        return true;
    }

    public BitSet not() {
        BitSet s = this.clone();
        s.notInPlace();
        return s;
    }

    public void notInPlace() {
        for (int i = bits.length - 1; i >= 0; i--) {
            bits[i] = ~bits[i];
        }
    }

    /** complement bits in the range 0..maxBit. */
    public void notInPlace(int maxBit) {
        notInPlace(0, maxBit);
    }

    /** complement bits in the range minBit..maxBit.*/
    public void notInPlace(int minBit, int maxBit) {
        // make sure that we have room for maxBit
        growToInclude(maxBit);
        for (int i = minBit; i <= maxBit; i++) {
            int n = wordNumber(i);
            bits[n] ^= bitMask(i);
        }
    }

    private final int numWordsToHold(int el) {
        return (el >> LOG_BITS) + 1;
    }

    public static BitSet of(int el) {
        BitSet s = new BitSet(el + 1);
        s.add(el);
        return s;
    }

    /** return this | a in a new set */
    public BitSet or(BitSet a) {
        BitSet s = this.clone();
        s.orInPlace(a);
        return s;
    }

    public void orInPlace(BitSet a) {
        // If this is smaller than a, grow this first
        if (a.bits.length > bits.length) {
            setSize(a.bits.length);
        }
        int min = Math.min(bits.length, a.bits.length);
        for (int i = min - 1; i >= 0; i--) {
            bits[i] |= a.bits[i];
        }
    }

    // remove this element from this set
    public void remove(int el) {
        int n = wordNumber(el);
        if (n >= bits.length) {
            growToInclude(el);
        }
        bits[n] &= ~bitMask(el);
    }

    /**
     * Sets the size of a set.
     * @param nwords how many words the new set should be
     */
    private void setSize(int nwords) {
        /*
        long newbits[] = new long[nwords];
        int n = Math.min(nwords, bits.length);
        System.arraycopy(bits, 0, newbits, 0, n);
        bits = newbits;
        */
        bits.length = nwords;        
    }

    public int size() {
        return bits.length << LOG_BITS; // num words * bits per word
    }

    /** return how much space is being used by the bits array not
     *  how many actually have member bits on.
     */
    public int lengthInLongWords() {
        return bits.length;
    }

    /**Is this contained within a? */
    public boolean subset(BitSet a) {
        if (a is null) return false;
        return this.and(a).equals(this);
    }

    /**Subtract the elements of 'a' from 'this' in-place.
     * Basically, just turn off all bits of 'this' that are in 'a'.
     */
    public void subtractInPlace(BitSet a) {
        if (a is null) return;
        // for all words of 'a', turn off corresponding bits of 'this'
        for (int i = 0; i < bits.length && i < a.bits.length; i++) {
            bits[i] &= ~a.bits[i];
        }
    }

    public int[] toArray() {
        int[] elems = new int[degree()];
        int en = 0;
        for (int i = 0; i < (bits.length << LOG_BITS); i++) {
            if (member(i)) {
                elems[en++] = i;
            }
        }
        return elems;
    }

    public long[] toPackedArray() {
        return bits;
    }

    public WString toWString() {
        return toWString(asWString(","));
    }

    /** Transform a bit set into a string by formatting each element as an integer
     * @separator The string to put in between elements
     * @return A commma-separated list of values
     */
    public WString toWString(WString separator) {
        WString str = new WString();
        for (int i = 0; i < (bits.length << LOG_BITS); i++) {
            if (member(i)) {
                if (str.length > 0) {
                    appS(str, separator);
                }
                appS(str, asWString(i));
            }
        }
        return str;
    }

    /** Transform a bit set into a string of characters.
     * @separator The string to put in between elements
     * @param formatter An object implementing the CharFormatter interface.
     * @return A commma-separated list of character constants.
     */
    public WString toWString(WString separator, CharFormatter formatter) {
        WString str = new WString();

        for (int i = 0; i < (bits.length << LOG_BITS); i++) {
            if (member(i)) {
                if (str.length > 0) {
                    appS(str, separator);
                }
                appS(str, formatter.literalChar(i));
            }
        }
        return str;
    }

    
    /**Create a string representation where instead of integer elements, the
     * ith element of vocabulary is displayed instead.  Vocabulary is a Vector
     * of Strings.
     * @param separator The string to put in between elements
     * @return A commma-separated list of character constants.
     */
    public WString toWString(WString separator, Vector!(WString) vocabulary) {
        if (vocabulary is null) {
            return toWString(separator);
        }
        WString str = new WString();
        for (int i = 0; i < (bits.length << LOG_BITS); i++) {
            if (member(i)) {
                if (str.length > 0) {
                    appS(str, separator);
                }
                if (i >= vocabulary.length()) {
                    appS(str, "<bad element " ~ asWString(i) ~ ">");
                }
                else if (vocabulary.elementAt(i) is null) {
                    appS(str, "<" ~ asWString(i) ~ ">");
                }
                else {
                    appS(str, vocabulary.elementAt(i));
                }
            }
        }
        return str;
    }

    /**
     * Dump a comma-separated list of the words making up the bit set.
     * Split each 64 bit number into two more manageable 32 bit numbers.
     * This generates a comma-separated list of C++-like unsigned long constants.
     */
    public WString toStringOfHalfWords() {
        WString s; // = new String();
        for (int i = 0; i < bits.length; i++) {
            if (i != 0)
                appS(s, ", ");
            long tmp = bits[i];
            tmp &= 0xFFFFFFFFL;
            appS(s, (asWString(tmp) ~ "UL"));
            appS(s, ", ");
            tmp = bits[i] >>> 32;
            tmp &= 0xFFFFFFFFL;
            appS(s, (asWString(tmp) ~ "UL"));
        }
        return s;
    }

    /**
     * Dump a comma-separated list of the words making up the bit set.
     * This generates a comma-separated list of Java-like long int constants.
     */
    public WString toStringOfWords() {
        WString s; // = new String();
        for (int i = 0; i < bits.length; i++) {
            if (i != 0)
                appS(s, ", ");
            appS(s, asWString(bits[i]) ~ "L");
        }
        return s;
    }

    /** Print out the bit set but collapse char ranges. */
    public WString toStringWithRanges(WString separator, CharFormatter formatter) {
        WString str = new WString();
        int[] elems = this.toArray();
        if (elems.length == 0) {
            return asWString("");
        }
        // look for ranges
        int i = 0;
        while (i < elems.length) {
            int lastInRange;
            lastInRange = 0;
            for (int j = i + 1; j < elems.length; j++) {
                if (elems[j] != elems[j - 1] + 1) {
                    break;
                }
                lastInRange = j;
            }
            // found a range
            if (str.length > 0) {
                appS(str, separator);
            }
            if (lastInRange - i >= 2) {
                appS(str, formatter.literalChar(elems[i]));
                appS(str, "..");
                appS(str, formatter.literalChar(elems[lastInRange]));
                i = lastInRange;    // skip past end of range for next range
            }
            else {    // no range, just print current char and move on
                appS(str, formatter.literalChar(elems[i]));
            }
            i++;
        }
        return str;
    }

    private final static int wordNumber(int abit) {
        return abit >> LOG_BITS; // abit / BITS
    }
}


/** Implementation of a StringBuffer-like object that does not have the
  unfortunate side-effect of creating Strings with very large buffers.
*/
public class ANTLRStringBuffer {
    protected wchar[] buffer = null;
    protected int reallength = 0;       // length and also where to store next char


    public this() {
        buffer = new wchar[50];
    }

    public this(int n) {
        buffer = new wchar[n];
    }

    public final void append(wchar c) {
        // This would normally be  an "ensureCapacity" method, but inlined
        // here for speed.
        if (reallength >= buffer.length) {
            // Compute a new length that is at least double old length
            int newSize = buffer.length;
            while (reallength >= newSize) {
                newSize *= 2;
            }
            buffer.length = newSize;
            
            // Allocate new array and copy buffer
            /*
            wchar[] newBuffer = new wchar[newSize];
            for (int i = 0; i < length; i++) {
                newBuffer[i] = buffer[i];
            }
            
            buffer = newBuffer;
            */
            
        }
        buffer[reallength] = c;
        reallength++;
    }

    public final void append(WString s) {
        for (int i = 0; i < s.length(); i++) {
            append(s.charAt(i));
        }
    }
    
    /// For better D compatibility
    public final void append(wchar[] s) {
        for (int i = 0; i < s.length; i++) {
            append(s[i]);
        }
    }
    
    /** For D compatibility
    */
    public final void appendS(char[] s) {
        append(asWString(s));
    }

    
    public final wchar charAt(int index) {
        return buffer[index];
    }

    final public wchar[] getBuffer() {
        return buffer;
    }

    public final int length() {
        return reallength;
    }

    public final void setCharAt(int index, wchar ch) {
        buffer[index] = ch;
    }

    public final void setLength(int newLength) {
        if (newLength < reallength) {
            reallength = newLength;
        }
        else {
            while (newLength > reallength) {
                append('\0');
            }
        }
    }

    public final WString toWString() {
        return new WString(buffer[0 .. reallength]);
    }
}



/// Using ANTLRStringBuffer as Java StringBuffer
alias ANTLRStringBuffer StringBuffer;



public abstract class FileLineFormatter {

    private static FileLineFormatter formatter;   // = new DefaultFileLineFormatter();
    static this()
    {
        formatter = new DefaultFileLineFormatter();
    }

    public static FileLineFormatter getFormatter() {
        return formatter;
    }

    public static void setFormatter(FileLineFormatter f) {
        formatter = f;
    }

    /** @param fileName the file that should appear in the prefix. (or null)
     * @param line the line (or -1)
     * @param column the column (or -1)
     */
    public abstract WString getFormatString(WString fileName, int line, int column);
}


public class DefaultFileLineFormatter : FileLineFormatter {
    public WString getFormatString(WString fileName, int line, int column) {
        StringBuffer buf = new StringBuffer();

        if (fileName != null)
            buf.append(fileName ~ ":");

        if (line != -1) {
            if (fileName == null)
                buf.append("line ");

            buf.append(asWString(line));

            if (column != -1)
                buf.append(":" ~ asWString(column));

            buf.append(":");
        }

        buf.append(" ");

        return buf.toWString();
    }
}



public struct StringUtils {
    /** General-purpose utility function for removing
     * characters from back of string
     * @param s The string to process
     * @param c The character to remove
     * @return The resulting string
     */
    static public WString stripBack(WString s, wchar c) {
        while (s.length > 0 && s.charAt(s.length - 1) == c) {
            s = s.substring(0, s.length - 1);
        }
        return s;
    }

    /** General-purpose utility function for removing
     * characters from back of string
     * @param s The string to process
     * @param remove A string containing the set of characters to remove
     * @return The resulting string
     */
    static public WString stripBack(WString s, WString remove) {
        boolean changed;
        do {
            changed = false;
            for (int i = 0; i < remove.length; i++) {
                wchar c = remove.charAt(i);
                while (s.length > 0 && s.charAt(s.length - 1) == c) {
                    changed = true;
                    s = s.substring(0, s.length - 1);
                }
            }
        } while (changed);
        return s;
    }

    /** General-purpose utility function for removing
     * characters from front of string
     * @param s The string to process
     * @param c The character to remove
     * @return The resulting string
     */
    static public WString stripFront(WString s, wchar c) {
        while (s.length > 0 && s.charAt(0) == c) {
            s = s.substring(1);
        }
        return s;
    }

    /** General-purpose utility function for removing
     * characters from front of string
     * @param s The string to process
     * @param remove A string containing the set of characters to remove
     * @return The resulting string
     */
    static public WString stripFront(WString s, WString remove) {
        boolean changed;
        do {
            changed = false;
            for (int i = 0; i < remove.length; i++) {
                wchar c = remove.charAt(i);
                while (s.length > 0 && s.charAt(0) == c) {
                    changed = true;
                    s = s.substring(1);
                }
            }
        } while (changed);
        return s;
    }

    /** General-purpose utility function for removing
     * characters from the front and back of string
     * @param s The string to process
     * @param head exact string to strip from head
     * @param tail exact string to strip from tail
     * @return The resulting string
     */
    public static WString stripFrontBack(WString src, WString head, WString tail) {
        int h = src.indexOf(head);
        int t = src.lastIndexOf(tail);
        if (h == -1 || t == -1) return src;
        return src.substring(h + 1, t);
    }
}
