/** Various utilities for the modules (including some
  faked java classes)
*/
module antlrrt.utils;

private import std.stream;
private import std.string;
private import std.stdio;

// private import antlrrt.exceptions;

alias int boolean;

/* Be careful: Java String objects are immutable, wchar[] is not.
*/
//alias wchar[] String;  // TODO

/** Java fake class: immutable String, renamed to WString
*/
class WString
{
    protected wchar[] data;
    protected static const int prime = 151;
    
    /** Convert wchar[] to char[]
    */
    // TODO: Better conversion for different charsets and UTF-8
    //     Avoid unnecessary copying
    
    this()
    {}
    
    this(wchar[] d)
    {
        data = d.dup;
    }
    
    this(char[] d)
    {
        data = toWChars(d);
    }
    
    this(wchar[] d, int start, int len)
    {
        data = d[start .. start + len].dup;
    }

    
    static char[] toString(wchar[] ws)
    {
        char[] result = new char[ws.length];
        for(int i = 0; i < ws.length; ++i)
            result[i] = cast(char) ws[i];
            
        return result;
    }
    
    static wchar[] toWChars(char[] s)
    {
        wchar[] result = new wchar[s.length];
        for(int i = 0; i < s.length; ++i)
            result[i] = cast(wchar) s[i];
            
        return result;
    }
    
    wchar[] toWChars()
    {
        return data.dup;
    }
    
    static WString toWString(char[] s)
    {
        return new WString(toWChars(s));
    }
            
    /**  This is the D version, not the Java version
    */
    char[] toString()
    {
        return toString(data);
    }
    
    boolean equalsIgnoreCase(WString comp)
    {
        return icmp(toString(), comp.toString()) == 0;  // TODO: Better conversion?
    }
    
    wchar charAt(int i)
    {
        return data[i];
    }
    
    int length()
    {
        return data.length;
    }
    
    WString substring(int begin, int end)
    {
        return new WString(data[begin .. end]);
    }
    
    WString substring(int begin)
    {
        return substring(begin, length());
    }
    
    int indexOf(WString f)
    {
        return find(toString(), f.toString());  // TODO: Better conversion?
    }
    
    int lastIndexOf(WString f)
    {
        return rfind(toString(), f.toString());  // TODO: Better conversion?
    }
    
    WString opCat(WString ws)
    {
        return new WString(data ~ ws.data);
    }
    
    /*WString opCat(char[] s)
    {
        return opCat(toWChars(s));
    }*/
    
    WString opCat(wchar[] w)
    {
        return new WString(data ~ w);
    }
    
    WString opCat(wchar c)
    {
        wchar[1] wa;
        wa[0] = c;
        return new WString(data ~ wa);
    }
    
    /*WString opCat_r(char[] s)
    {
        return opCat_r(toWChars(s));
    }*/
    
    WString opCat_r(wchar[] w)
    {
        return new WString(w ~ data);
    }
    
    WString opCat_r(wchar c)
    {
        wchar[1] wa;
        wa[0] = c;
        return new WString(wa ~ data);
    }
    
    /// D hash function
    uint toHash() {
        return cast(uint) hashCode();
    }
    
    int opCmp(Object o) {
        WString s = cast(WString) o;
        
        if (s is null)
            return super.opCmp(o);
        
        int l = length();
        if (l != s.length()) {
            return l - s.length();
        }
        
        for (int i = 0; i < l; i++) {
            if (data[i] != s.data[i]) {
                return data[i] - s.data[i];
            }
        }
        
        return 0;
    }
    
    /// From Java
    int hashCode() {
        
        int hashval = 0;
        int l = length();

        for (int i = 0; i < l; i++) {
            hashval = hashval * prime + charAt(i);
        }

        return hashval;
    }
}


  
template appendWST(T)
{
    /** Template for appS functions to replace Java:
        <code>
        String w = new String();
        
        w += "Something";
        </code>
        
        by D code:    
        <code>
        WString w = new WString();
        
        appS(w, "Something");
        </code>
    */ 
    void appendWST(inout WString w, T a)
    {
        w = w ~ a;
    }
}

alias appendWST!(WString) appS;
// alias appendWST!(char[]) appS;
alias appendWST!(wchar[]) appS;
alias appendWST!(wchar) appS;



alias std.stream.InputStream InputStream;


template asWStringT(T)
{
    WString asWStringT(T e)
    {
        return new WString(std.string.toString(e));
    }
}


alias asWStringT!(int) asWString;
alias asWStringT!(long) asWString;

WString asWString(char[] e)
{
    return new WString(e);
}

WString asWString(wchar e)
{
    wchar[] result = new wchar[1];
    result[0] = e;
    
    return new WString(result);
}


/// Convert an array of wchar[][] to WString[]
WString[] asWStringArr(wchar[][] arr)
{
    WString[] result = new WString[arr.length];
    for(int i = 0; i < arr.length; ++i)
        result[i] = new WString(arr[i]);

    return result;
}




/** Return an int converted to an upper case hex string.
*/
WString asUpperHexWString(int i)
{
    return new WString(toupper(toString(cast(long)i, 16u)));
}


/** Java fake class. A Java Reader provides read access to a stream of
    characters, normally by converting from an underlying stream of bytes.
*/
interface Reader
{
    wchar read();
}

/** Very basic implementation for Reader. Simply casts each char to a wchar
  Much TODO.
*/
class BinaryReader : Reader
{
    InputStream mInp;
    
    this(InputStream inp)
    {
        mInp = inp;
    }
    
    wchar read()
    {
        char c = mInp.getc();
        if (c == char.init && mInp.eof)
            return wchar.init;
            
        return cast(wchar)c;
    }
}    


/** Very basic implementation for Reader. Much TODO.
*/
alias BinaryReader StreamReader;
    

/** Java fake class. A Java Writer provides write access to a stream of
    characters, normally by converting to an underlying stream of bytes.
*/
interface Writer
{
    void write(WString s);
}


/** Very basic implementation for Writer. Much TODO.
*/
class StreamWriter : Writer
{
    OutputStream mOutp;
    
    this(OutputStream outp)
    {
        mOutp = outp;
    }
    
    void write(WString ws)
    {
        char[] s = ws.toString();
        foreach(char c; s)
            mOutp.write(c);  // The format of write(char[]) is not specified
    }
}    



/** Java fake class
*/
struct System
{
    // TODO Support writing to stderr
    static struct Stream
    {
        void print(WString s)
        {
            writef("%s", s.toString());   // TODO: Conversion needed?
        }            
 
        void println(WString s)
        {
            writefln("%s", s.toString());
        }  
           
        void print(wchar[] s)
        {
            writef("%s", WString.toString(s));   // TODO: Conversion needed?
        }            
 
        void println(wchar[] s)
        {
            writefln("%s", WString.toString(s));
        }     
    }
    
    static Stream outp;  // In Java "out", but this is reserved in D
    static Stream err;
    
    static void exit(int c)
    {
        writef("Exit %i", c);   // TODO !!!
    }
}

/** Java fake class
*/
struct Math
{
    static int min(int a, int b)
    {
        return a < b ? a : b;
    }
    
    static int max(int a, int b)
    {
        return a > b ? a : b;
    }
}

/** Java fake class
*/
struct Character
{
    private import std.ctype;

    static wchar toLowerCase(wchar c)
    {
        return std.ctype.tolower(c);
    }
}



public class ANTLRException : Exception {

    protected WString msg;
    
    public this() {
        super("");
    }

    public this(WString s) {
        super(s.toString());
        msg = s;
    }
    
    public WString getMessage() {
        return msg;
    }
    
    public char[] toString() {
        return getMessage().toString();
    }
}


/**
 * For ASTEnumeration
 */
public class NoSuchElementException : ANTLRException {
    /**
     * NoSuchElementException constructor comment.
     * @param s WString
     */
    public this(WString s) {
        super(s);
    }
}


public class IllegalArgumentException : ANTLRException {
    /**
     * IllegalArgumentException constructor comment.
     * @param s WString
     */
    public this(WString s) {
        super(s);
    }
}

public class RuntimeException : ANTLRException {
    /**
     * RuntimeException constructor comment.
     * @param s WString
     */
    public this(WString s) {
        super(s);
    }
}

alias std.stream.StreamException IOException;
